"""Tests for interface class."""

from typing import Any

from aiohttp import ClientSession, CookieJar
from aiohttp.test_utils import TestClient, TestServer, loop_context
import pytest

from aioptdevices.configuration import Configuration
from aioptdevices.errors import (
    PTDevicesForbiddenError,
    PTDevicesRequestError,
    PTDevicesUnauthorizedError,
)
from aioptdevices.interface import Interface, PTDevicesResponse

# Add your token and device ID, **keep out of git**
from .mock_api import (
    API_URL,
    BAD_GATEWAY_ERROR_DEVICE_ID,
    FORBIDDEN_ERROR_DEVICE_ID,
    GOOD_RESP_DICT,
    NORMAL_DEVICE_ID,
    NORMAL_DEVICE_MAC,
    TOKEN,
    UNAUTHORIZED_ERROR_DEVICE_ID,
    WRONG_CONTENT_TYPE_DEVICE_ID,
)
from .secret import DEVICE_ID as secret_DEVICE_ID, TOKEN as secret_TOKEN

# Query Paramaters


def test_interface(test_web_app):
    """Test the basic interface connection."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)
        loop.run_until_complete(client.start_server())

        normal_config: Configuration = Configuration(
            auth_token=TOKEN,
            device_id=NORMAL_DEVICE_ID,
            url=str(client.make_url(API_URL)),
            session=client.session,
        )
        interface: Interface = Interface(normal_config)

        async def test_get_data():

            resp: PTDevicesResponse = await interface.get_data()
            assert resp.get("body") == GOOD_RESP_DICT.get("data")

        loop.run_until_complete(test_get_data())
        loop.run_until_complete(client.close())


def test_mac_id(test_web_app):
    """Test the basic interface connection."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)
        loop.run_until_complete(client.start_server())

        normal_config: Configuration = Configuration(
            auth_token=TOKEN,
            device_id=NORMAL_DEVICE_MAC,
            url=str(client.make_url(API_URL)),
            session=client.session,
        )
        interface: Interface = Interface(normal_config)

        async def test_get_data():

            resp: PTDevicesResponse = await interface.get_data()
            assert resp.get("body") == GOOD_RESP_DICT.get("data")

        loop.run_until_complete(test_get_data())
        loop.run_until_complete(client.close())


def test_resp_error_handling(test_web_app):
    """Test the interface with different response statuses."""
    with loop_context() as loop:
        # Setup the test server
        server = TestServer(test_web_app)
        client = TestClient(server, loop=loop)

        loop.run_until_complete(client.start_server())

        def make_interface(device_id: str) -> Interface:
            return Interface(
                Configuration(
                    auth_token=TOKEN,
                    device_id=device_id,
                    url=str(client.make_url(API_URL)),
                    session=client.session,
                )
            )

        async def test_errors():
            # UNAUTHORIZED (401) Handling
            interface: Interface = make_interface(UNAUTHORIZED_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesUnauthorizedError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # FORBIDDEN (403) Handling
            interface: Interface = make_interface(FORBIDDEN_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesForbiddenError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # PTDevicesRequestError (403) Handling
            interface: Interface = make_interface(BAD_GATEWAY_ERROR_DEVICE_ID)

            with pytest.raises(PTDevicesRequestError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # Wrong content type Handling
            interface: Interface = make_interface(WRONG_CONTENT_TYPE_DEVICE_ID)

            with pytest.raises(PTDevicesRequestError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

            # Bad URL handling
            interface: Interface = Interface(
                Configuration(
                    auth_token=TOKEN,
                    device_id=NORMAL_DEVICE_ID,
                    url=API_URL,
                    session=client.session,
                )
            )

            with pytest.raises(PTDevicesRequestError):  # Test UNAUTHORIZED (401)
                await interface.get_data()

        loop.run_until_complete(test_errors())
        loop.run_until_complete(client.close())


def test_real_server():
    """Test the interface with the real server."""
    with loop_context() as loop:
        # Setup the test server

        async def test_real_server():
            session: ClientSession = ClientSession(cookie_jar=CookieJar(unsafe=True))
            interface: Interface = Interface(
                Configuration(
                    auth_token=secret_TOKEN,
                    device_id=secret_DEVICE_ID,
                    url="http://www.ptdevices.com/token/v1/device/",
                    session=session,
                )
            )
            resp: PTDevicesResponse = await interface.get_data()
            data: dict[str, Any] | None = resp.get("body")
            assert data
            assert "device_id" in data
            await session.close()

        loop.run_until_complete(test_real_server())


# interface: Interface = Interface(normal_config)

# loop.run_until_complete(client.start_server())


# def test_configuration(test_web_app):
#     """Test the interface connection to a mock server."""
#     with loop_context() as loop:
#         server = TestServer(test_web_app)
#         client = TestClient(server, loop=loop)
#         loop.run_until_complete(client.start_server())

#         async def test_get_routes():
#             resp = await client.get(f"{API_URL}{NORMAL_DEVICE_ID}", params=query_params)

#             # resp = await client.get(f"{API_URL}{NORMAL_DEVICE_ID}?api_token={TOKEN}")
#             assert resp.status == 200
#             resp = await client.get(
#                 f"{API_URL}{BAD_GATEWAY_ERROR_DEVICE_ID}", params=query_params
#             )
#             assert resp.status == 502

#         loop.run_until_complete(test_get_routes())
#         loop.run_until_complete(client.close())

# t_serv = test_server()
# test_configuration: Configuration = Configuration(
#     auth_token=
#     device_id=

# )



"""Configuration Library."""

from dataclasses import dataclass

from aiohttp import ClientSession


@dataclass
class Configuration:
    """PTDevices Communication Configuration."""

    auth_token: str  # Used for authorizing communication with server
    device_id: str  # Used to identify the device to get data from
    url: str  # URL to get data from
    session: ClientSession  # Connection session with servers



"""CLI for PTDevices."""

import argparse
import asyncio
from asyncio.timeouts import timeout
import logging

from aiohttp import ClientSession, CookieJar

import aioptdevices
from aioptdevices.configuration import Configuration
from aioptdevices.interface import Interface

LOGGER = logging.getLogger(__name__)


async def connect(
    deviceID: str, authToken: str, url: str, webSession: ClientSession
) -> Interface | None:
    """Set up and Connect to PTDevices."""

    # Setup interface to PTDevices
    interface: Interface = Interface(
        Configuration(authToken, deviceID, url, webSession)
    )
    try:
        async with timeout(10):
            data = await interface.get_data()
            LOGGER.info("Data: %s", data.get("body"))
    except aioptdevices.PTDevicesRequestError as err:
        LOGGER.warning("failed to connect to PTDevices server: %s", err)

    except aioptdevices.PTDevicesUnauthorizedError as err:
        LOGGER.warning("Unable to read device data because of bad token: %s", err)

    except aioptdevices.PTDevicesForbiddenError as err:
        LOGGER.warning("Unable, device does not belong to the token holder: %s", err)
    else:
        return interface
    return None


async def main(
    deviceID: str,
    authToken: str,
    url: str,
) -> None:
    """Configure a connection and get the token api data from server."""

    # ----------  Connecting To PTDevices  -----------

    LOGGER.info(
        "\n%s\n", "  Connecting To PTDevices  ".center(48, "-")
    )  # Output a section title for connecting

    # Create a web session for use when connecting
    session: ClientSession = ClientSession(cookie_jar=CookieJar(unsafe=True))
    # session = ClientSession()

    # Setup connection to PTDevices
    ptdevicesInterface = await connect(deviceID, authToken, url, session)

    if not ptdevicesInterface:  # Failed connection to PTDevices
        LOGGER.error("Failed to connect to PTDevices")
        await session.close()
        return

    # -----------------  Connected  ------------------

    LOGGER.info(
        "\n%s\n", "  Connected  ".center(48, "-")
    )  # Output a section title when connected
    LOGGER.info("Successfully connected to %s", deviceID)

    # try:
    #     data = await ptdevicesInterface.get_data()
    #     LOGGER.info(f"Device {deviceID} data: \n{data}")
    # except aioptdevices.PTDevicesRequestError as err:
    #     LOGGER.warning(f"failed to connect to PTDevices server: {err}")

    # except aioptdevices.PTDevicesUnauthorizedError as err:
    #     LOGGER.warning(f"Unable, to read device data because of bad token: {err}")

    # except aioptdevices.PTDevicesForbiddenError as err:
    #     LOGGER.warning(f"Unable, device does not belong to the token holder: {err}")

    await session.close()
    return


def starter():
    """Parse CLI Arguments and fetch device data."""
    default_url = "https://www.ptdevices.com/token/v1/device/"

    # Parse cli args
    parser = argparse.ArgumentParser()
    parser.add_argument("deviceID", type=int)
    parser.add_argument("authToken", type=str)
    parser.add_argument("-U", "--url", type=str, default=default_url)
    parser.add_argument("-D", "--debug", action="store_true")
    args = parser.parse_args()

    # Set the log level
    LOG_LEVEL = logging.INFO
    if args.debug:
        LOG_LEVEL = logging.DEBUG

    logging.basicConfig(format="%(message)s", level=LOG_LEVEL)

    # Output the settings
    # LOGGER.info(
    #     f"deviceID: {args.deviceID}\nToken: {args.authToken}\nurl: {args.url}\ndebug: {args.debug}\n"
    # )

    # --------------------  ARGS  --------------------

    LOGGER.info("\n%s\n", "  ARGS  ".center(48, "-"))  # Output a section title for args
    LOGGER.info("deviceID: %s", args.deviceID)
    # LOGGER.info("Token: %s", args.authToken)
    LOGGER.info("url: %s", args.url)
    LOGGER.info("debug: %s", args.debug)

    # Run the program
    try:
        asyncio.run(
            main(
                deviceID=args.deviceID,
                authToken=args.authToken,
                url=args.url,
            )
        )

    except KeyboardInterrupt:
        LOGGER.info("Keyboard interrupt")


if __name__ == "__main__":
    starter()
# ----------------------  ARGS  ------------------



"""Python Library for communicating with PTDevices."""

from http import HTTPStatus
import logging
from typing import Any, TypedDict

import orjson

from aioptdevices.errors import (
    PTDevicesForbiddenError,
    PTDevicesRequestError,
    PTDevicesUnauthorizedError,
)

from .configuration import Configuration

LOGGER = logging.getLogger(__name__)


class PTDevicesResponse(TypedDict, total=False):
    """Typed Response from PTDevices."""

    body: dict[str, Any]
    code: int


class Interface:
    """Interface for PTDevices."""

    def __init__(self, config: Configuration) -> None:
        """Initilize object variables."""
        self.config = config

    async def get_data(self) -> PTDevicesResponse:
        """Fetch device data from PTDevices server and format it."""
        # Request url: https://api.ptdevices.com/token/v1/device/{deviceId}?api_token={given_token}
        # Where
        #   {deviceId} is the numeric internal device id,
        #       found in the url https://www.ptdevices.com/device/level/{deviceId}
        #   {given_token} is the access token you were given

        url = f"{self.config.url}{self.config.device_id}?api_token={self.config.auth_token}"

        LOGGER.debug(
            "Sending request to %s for data from device #%s",
            self.config.url,
            self.config.device_id,
        )

        async with self.config.session.request(
            "get",
            url,
            allow_redirects=False,
        ) as results:
            LOGGER.debug(
                "%s Received from %s",
                results.status,
                self.config.url,
            )

            # Check return code
            if results.status == HTTPStatus.UNAUTHORIZED:  # 401
                raise PTDevicesUnauthorizedError(
                    f"Request to {url.split('?api_token')[0]} failed, the token provided is not valid"
                )
            elif results.status == HTTPStatus.FOUND:  # 302
                # Back end currently returns a 302 when request is not authorized
                raise PTDevicesUnauthorizedError(
                    f"Request to {url.split('?api_token')[0]} failed, the token provided is not valid (302)"
                )

            elif results.status == HTTPStatus.FORBIDDEN:  # 403
                raise PTDevicesForbiddenError(
                    f"Request to {url.split('?api_token')[0]} failed, token invalid for device {self.config.device_id}"
                )

            elif results.status != HTTPStatus.OK:  # anything but 200
                raise PTDevicesRequestError(
                    f"Request to {url.split('?api_token')[0]} failed, got unexpected response from server ({results.status})"
                )

            # Check content type
            elif (
                results.content_type != "application/json"
                or results.content_length == 0
            ):
                raise PTDevicesRequestError(
                    f"Failed to get device data, returned content is invalid. Type: {results.content_type}, content Length: {results.content_length}, content: {results.content}"
                )

            raw_json = await results.read()

            body = orjson.loads(raw_json)

            return PTDevicesResponse(
                code=results.status,
                body=body["data"],
            )



"""AioPTDevices Errors."""


class PTDevicesError(Exception):
    """Base Error for PTDevices."""


class PTDevicesRequestError(PTDevicesError):
    """The request was unable to be fulfilled.

    Caused by a bad url
    """


class PTDevicesUnauthorizedError(PTDevicesError):
    """The request was denied because the user does not have permission to access the device requested.

    The server did not recognize the token
    """


class PTDevicesForbiddenError(PTDevicesError):
    """The request was denied because the user does not have permission to access the device requested.

    The server recognized the token but the device is not accessible by the linked account
    """
